/* ========================================================================
 * Copyright (c) 2009 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */

package edu.internet2.middleware.subject.provider;

import java.util.Properties;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collections;
import java.util.Comparator;
import java.util.TreeSet;

import javax.net.ssl.SSLSocketFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import edu.internet2.middleware.subject.SourceUnavailableException;
import edu.internet2.middleware.subject.Subject;
import edu.internet2.middleware.subject.SubjectNotFoundException;
import edu.internet2.middleware.subject.SubjectNotUniqueException;
import edu.internet2.middleware.subject.SubjectType;

import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.SearchResult;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;

import edu.vt.middleware.ldap.Ldap;
import edu.vt.middleware.ldap.LdapConfig;
import edu.vt.middleware.ldap.pool.SoftLimitLdapPool;
import edu.vt.middleware.ldap.pool.DefaultLdapFactory;
import edu.vt.middleware.ldap.pool.LdapPoolConfig;
import edu.vt.middleware.ldap.LdapUtil;

/**
 * Ldap source adapter.  Configuration is from a properties file
 */

public class LdapSourceAdapter extends BaseSourceAdapter {
    
    private static Log log = LogFactory.getLog(LdapSourceAdapter.class);
    
    private String nameAttributeName = null;
    private String subjectIDAttributeName = null;
    private String descriptionAttributeName = null;
    private String subjectTypeString = null;
    private String localDomain = null;
    private String propertiesFile = null;
    private SoftLimitLdapPool ldapPool;
    private boolean initialized = false;

    public LdapSourceAdapter() {
        super();
    }
    
    public LdapSourceAdapter(String id, String name) {
        super(id, name);
    }
    
    // initialization

    /**
     * {@inheritDoc}
     */
    public void init() {
        log.debug("ldap source init");
        Properties props = getInitParams();

	nameAttributeName = getStringProperty(props,"Name_AttributeType");
	subjectIDAttributeName = getStringProperty(props,"SubjectID_AttributeType");
	descriptionAttributeName = getStringProperty(props,"Description_AttributeType");

        propertiesFile = getStringProperty(props,"ldapProperties_file");
        initializeLdap();
   }

   private void initializeLdap() {
   
        log.debug("ldap initializeLdap");
        
        // all ldap config from the ldap properties file
        LdapConfig ldapConfig = LdapConfig.createFromProperties(propertiesFile);

        // including config for the pem cert mode
        Map<String, Object> props = ldapConfig.getEnvironmentProperties();
        String cafile = (String)props.get("pemCaFile");
        String certfile = (String)props.get("pemCertFile");
        String keyfile = (String)props.get("pemKeyFile");
        if (cafile!=null && certfile!=null && keyfile!=null) {
            log.debug("using the PEM socketfactory: ca=" + cafile + ", cert=" + certfile + ", key=" + keyfile);
            LdapPEMSocketFactory sf = new LdapPEMSocketFactory(cafile, certfile, keyfile);
            SSLSocketFactory ldapSocketFactory = sf.getSocketFactory();
            ldapConfig.setSslSocketFactory(ldapSocketFactory);
        } else {
            log.debug("using the default socketfactory");
        }

	DefaultLdapFactory factory = new DefaultLdapFactory(ldapConfig);
	LdapPoolConfig poolConfig = new LdapPoolConfig();
     
        try {
	   ldapPool = new SoftLimitLdapPool(factory);
           ldapPool.initialize();
           initialized = true;
        } catch (Exception e) {
           log.debug("ldappool error = " + e);
        }
        log.debug("ldap initialize done");
  }
    

    // Public methods

    /**
     * {@inheritDoc}
     */
    public Subject getSubject(String id) throws SubjectNotFoundException,SubjectNotUniqueException {
        Subject subject = null;
        Search search = getSearch("searchSubject");
        if (search == null) {
            log.error("searchType: \"searchSubject\" not defined.");
            return subject;
        }
        String[] attributeNames = {nameAttributeName, descriptionAttributeName, subjectIDAttributeName};
        Attributes attributes = getLdapUnique( search, id, attributeNames   );
        subject = createSubject(attributes);
        if (subject == null) {
            throw new SubjectNotFoundException("Subject " + id + " not found.");
        }
        return subject;
    }
    
    /**
     * {@inheritDoc}
     */
    public Subject getSubjectByIdentifier(String id) throws SubjectNotFoundException,SubjectNotUniqueException {
        Subject subject = null;
        Search search = getSearch("searchSubjectByIdentifier");
        if (search == null) {
            log.error("searchType: \"searchSubjectByIdentifier\" not defined.");
            return subject;
        }
        String[] attributeNames = {nameAttributeName, subjectIDAttributeName, descriptionAttributeName};
        if (localDomain != null) {
           int atpos;
           if ((atpos=id.indexOf("@" + localDomain))>0) {
              log.debug("looking at id=" + id);
              id = id.substring(0, atpos);
              log.debug("converted to id=" + id);
           }
        }
        Attributes attributes = getLdapUnique( search, id, attributeNames );
        subject = createSubject(attributes);
        if (subject == null) {
            throw new SubjectNotFoundException("Subject " + id + " not found.");
        }
        return subject;
    }
    
    /**
     * {@inheritDoc}
     */
    public Set search(String searchValue) {
        Comparator cp = new LdapComparator();
        TreeSet result = new TreeSet(cp);
        Search search = getSearch("search");
        if (search == null) {
            log.error("searchType: \"search\" not defined.");
            return result;
        }
        String[] attributeNames = {nameAttributeName, subjectIDAttributeName,descriptionAttributeName} ;
        Iterator<SearchResult> ldapResults = getLdapResults(search,searchValue, attributeNames);
        if (ldapResults == null) {
            return result;
        }
        while ( ldapResults.hasNext()) {
            SearchResult si = (SearchResult) ldapResults.next();
            Attributes attributes = si.getAttributes();
            Subject subject = createSubject(attributes);
            result.add(subject);
        }

        // sort the result
        log.debug("set has " + result.size() + " subjects");
        if (result.size()>0) log.debug("first is " + ((Subject)result.first()).getDescription());
        return result;
    }
    
    private Subject createSubject( Attributes attributes) {
        String name = "";
        String subjectID = "";
        String description = "";

        if (attributes==null) {
           log.debug("ldap createsubject with null attrs");
           return (null);
        }
        try {
            Attribute attribute = attributes.get(subjectIDAttributeName);
            if (attribute == null) {
                log.error("No value for LDAP attribute \"" + subjectIDAttributeName + "\". It is Grouper attribute \"SubjectID\".");
                return null;
            }
            subjectID   = (String) attribute.get();
            attribute = attributes.get(nameAttributeName);
            if (attribute == null) {
                log.error("No value for LDAP attribute \"" + nameAttributeName + "\". It is Grouper attribute \"name\".");
                return null;
            }
            name = (String) attribute.get();
            attribute = attributes.get(descriptionAttributeName);
            if (attribute == null) {
                log.error("No value for LDAP attribute \"" + descriptionAttributeName + "\". It is Grouper attribute \"description\".");
            } else {
                description   = (String) attribute.get();
            }
        } catch (NamingException ex) {
            log.error("LDAP Naming Except: " + ex.getMessage(), ex);
        }
        return new LdapSubject(subjectID,  name, description, this.getSubjectType(), this);
    }
    
    protected String getStringProperty(Properties props, String prop) {
        String value = props.getProperty(prop);
        if (value==null) {
            log.error("Property '" + prop + "' is not defined!");
        }
        return (value);
    }

    
    /**
     * Loads attributes for the argument subject.
     */
    protected Map loadAttributes(LdapSubject subject) {
        Map attributes = new HashMap();
        Search search = getSearch("searchSubject");
        if (search == null) {
            log.error("searchType: \"search\" not defined.");
            return attributes;
        }

        //setting attributeNames to null will cause all attributes for a subject to be returned
        String[] attributeNames = null;
        Set attributeNameSet = this.getAttributes();
        if (attributeNameSet.size()==0) {
            attributeNames = null;
        } else {
            attributeNames = new String[attributeNameSet.size()];
            int i = 0;
            for (Iterator it = attributeNameSet.iterator(); it.hasNext(); attributeNames[i++]= (String) it.next());
        }
        try {
            Attributes ldapAttributes = getLdapUnique(search,subject.getId(), attributeNames);
            for (NamingEnumeration e = ldapAttributes.getAll(); e.hasMore();) {
                Attribute attr = (Attribute) e.next();
                String name = attr.getID();
                Set values = new HashSet();
                for (NamingEnumeration en = attr.getAll(); en.hasMore(); ) {
                    Object value =  en.next();
                    values.add(value.toString());
                }
                attributes.put(name, values);
            }
            subject.setAttributes(attributes);
        } catch (SubjectNotFoundException ex ) {
            log.error("SubjectNotFound: "+ subject.getId() +" " + ex.getMessage(), ex);
        } catch (SubjectNotUniqueException ex ) {
            log.error("SubjectNotUnique: "+ subject.getId() +" " + ex.getMessage(), ex);
        } catch (NamingException ex ) {
            log.error("LDAP Naming Except: " + ex.getMessage(), ex);
        }
        return attributes;
    }
    
   protected Iterator<SearchResult> getLdapResults(Search search, String searchValue, String[] attributeNames) {
        Ldap ldap = null;
        String filter = null;
        Iterator<SearchResult> results = null;
        int cp;

        if (!initialized) initializeLdap();

        if ((cp=searchValue.indexOf(',')) >0 ) {
            log.debug("first, last name search");
            filter = search.getParam("firstlastfilter");
            if (filter==null) {
                log.error("Search firstlastfilter not found for search type:  " + search.getSearchType());
                return results;
            }
            String last = searchValue.substring(0, cp);
            String first = searchValue.substring(cp+1);
            filter = filter.replaceAll("%LAST%", escapeSearchFilter(last));
            filter = filter.replaceAll("%FIRST%", escapeSearchFilter(first));
         } else {
            // simple search
            filter = search.getParam("filter");
            if (filter==null) {
                log.error("Search filter not found for search type:  " + search.getSearchType());
                return results;
            }
            filter = filter.replaceAll("%TERM%", escapeSearchFilter(searchValue));
        }
        log.debug("searchType: " + search.getSearchType() + " filter: " + filter);

        try  {
            ldap =  (Ldap) ldapPool.checkOut();
            results = ldap.search(filter, attributeNames );
        } catch (NamingException ex) {
            log.error("Ldap NamingException: " + ex.getMessage(), ex);
        } catch (Exception ex) {
            log.error("Ldap Exception: " + ex.getMessage(), ex);
            
        } finally {
            if (ldap != null) {
                try {
                    ldapPool.checkIn(ldap);
                } catch (Exception e) {
                    log.error("Could not return Ldap object back to pool", e);
                }
            }
        }
        return results;
    }
    
    
    protected Attributes getLdapUnique( Search search, String searchValue, String[] attributeNames)
                 throws SubjectNotFoundException,SubjectNotUniqueException  {
        Attributes attributes = null;
        Iterator<SearchResult> results = getLdapResults(search, searchValue, attributeNames);
        
        if (results == null || !results.hasNext()) {
            String errMsg = "No results: " + search.getSearchType() + " filter:" + search.getParam("filter") + " searchValue: " + searchValue;
            throw new SubjectNotFoundException( errMsg);
        }
            
        SearchResult si = ( SearchResult )results.next( );
        attributes = si.getAttributes();
        if ( results.hasNext()) {
            si = (SearchResult) results.next();
            String errMsg ="Search is not unique:" + si.getName() + "\n";
            throw new SubjectNotUniqueException( errMsg );
        }
        return attributes;
    }
    
    /**
     * Escape a search filter to prevent LDAP injection.
     * From http://www.owasp.org/index.php/Preventing_LDAP_Injection_in_Java
     * 
     * @param filter
     * @return escaped filter
     */
    protected String escapeSearchFilter(String filter) {
        //From RFC 2254
        String escapedStr = new String(filter);
        escapedStr = escapedStr.replaceAll("\\\\","\\\\5c");
        // We want people to be able to use wildcards.
        // escapedStr = escapedStr.replaceAll("\\*","\\\\2a");
        escapedStr = escapedStr.replaceAll("\\(","\\\\28");
        escapedStr = escapedStr.replaceAll("\\)","\\\\29");
        escapedStr = escapedStr.replaceAll("\\"+Character.toString('\u0000'), "\\\\00");
        return escapedStr;
    }
    /**
     * @see edu.internet2.middleware.subject.Source#checkConfig()
     */
    public void checkConfig() {
    }
    /**
     * @see edu.internet2.middleware.subject.Source#printConfig()
     */
    public String printConfig() {
      String message = "sources.xml ldap source id:   " + this.getId() + ": " + propertiesFile;
      return message;
    }
}
